#ifndef __ODE_COMMON_H_
#define __ODE_COMMON_H_

#ifdef WIN32
#include <windows.h>
#endif

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

 
#ifdef dDOUBLE
#define dsDrawBox dsDrawBoxD
#define dsDrawSphere dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCapsule dsDrawCapsuleD
#define dsDrawConvex dsDrawConvexD
#define dsDrawLine dsDrawLineD
#define dsDrawTriangle dsDrawTriangleD
#endif



struct ST_Object
{
    dGeomID geom;
    dBodyID body;
    float color[4];

    ST_Object() : geom(NULL), body(NULL) {
        color[0] = 1.0; color[1] = 1.0; color[2] = 1.0; color[3] = 1.0;
    }

    void setColor( float r, float g, float b, float a=1.0 ) {
        color[0] = r; color[1] = g; color[2] = b; color[3] = a;
    }
};


inline
void drawObj( const ST_Object *obj )
{
    if (!obj->geom) return;

    const dReal *pos = dGeomGetPosition( obj->geom );
    const dReal *R = dGeomGetRotation (obj->geom);

    dsSetColorAlpha( obj->color[0], obj->color[1], obj->color[2], obj->color[3] );

    int type = dGeomGetClass (obj->geom);
    if (type == dBoxClass) {
        dVector3 sides;
        dGeomBoxGetLengths (obj->geom,sides);
        dsDrawBox (pos,R,sides);
    }
    else if (type == dSphereClass) {
        dsDrawSphere (pos,R,dGeomSphereGetRadius(obj->geom));
    }
    else if (type == dCapsuleClass) {
        dReal radius,length;
        dGeomCapsuleGetParams (obj->geom,&radius,&length);
        dsDrawCapsule (pos,R,length,radius);
    }
    else if (type == dCylinderClass) {
        dReal radius,length;
        dGeomCylinderGetParams (obj->geom,&radius,&length);
        dsDrawCylinder (pos,R,length,radius);
    }
    else if (type == dRayClass) {
        dVector3 Origin, Direction;
        dGeomRayGet(obj->geom, Origin, Direction);
        dReal Length = dGeomRayGetLength(obj->geom);

        dVector3 End;
        End[0] = Origin[0] + (Direction[0] * Length);
        End[1] = Origin[1] + (Direction[1] * Length);
        End[2] = Origin[2] + (Direction[2] * Length);
        End[3] = Origin[3] + (Direction[3] * Length);

        dsDrawLine(Origin, End);
    }
    else if (type == dTriMeshClass) {
        int triCount = dGeomTriMeshGetTriangleCount( obj->geom );
        dReal pos_[3] = {0.0, 0.0, 0.0};
        for (int ii = 0; ii < triCount; ii++) {
            dVector3 v0, v1, v2;
            dGeomTriMeshGetTriangle( obj->geom, ii, &v0, &v1, &v2 );
            dsDrawTriangle( pos_, R, v0, v1, v2, 1 );
        }
    }
}

inline
void drawJoint( const ST_Object *obj )
{
    if (!obj->body) return;

    int num = dBodyGetNumJoints(obj->body);
    for (int i=0; i<num; i++) {
        dJointID joint = dBodyGetJoint( obj->body, i );
        int type = dJointGetType( joint );
        if (type == dJointTypeHinge) {
        
            dsSetColor( 1.0, 0.0, 0.0 );

            dVector3 Origin, Direction;
            dReal length = 2.0;
            dJointGetHingeAnchor( joint, Origin );
            dJointGetHingeAxis( joint, Direction );

            dVector3 End;
            End[0] = Origin[0] + (Direction[0] * length);
            End[1] = Origin[1] + (Direction[1] * length);
            End[2] = Origin[2] + (Direction[2] * length);
            End[3] = Origin[3] + (Direction[3] * length);

            dsDrawLine(Origin, End);
        }
    }
}


#endif