#include "ode_support.h"
#include <vector>

static dWorldID world;
static dSpaceID space;
static dJointGroupID contactgroup;

ST_Object* car_body = NULL;    // ԑ
ST_Object* car_wheel_f = NULL; // Oԗ
ST_Object* car_wheel_r = NULL; // Eԗ
ST_Object* car_wheel_l = NULL; // 덶ԗ
ST_Object* car_sensor = NULL;  // ZT[{fB
ST_Object* sensor_r = NULL;    // EZT[
ST_Object* sensor_l = NULL;    // ZT[
ST_Object* w_line = NULL;      // 
std::vector<ST_Object*> drawObjList;


int hit_l = 0;   // 1:ZTm, 0:mȂ
int hit_r = 0;   // 1:EZTm, 0:mȂ


////////////////////////////////////////////////////////////
// f[^@i1mj
struct WhiteLine {
    static const int VertexCount = 8;
    static const int IndexCount = 8 * 3;
    static const float Vertices[VertexCount * 3];

    static const dTriIndex Indices[IndexCount];
};
const float WhiteLine::Vertices[VertexCount * 3] = {
     0.0,  0.0, 0.0,  // 1
     1.0,  1.0, 0.0,  // 2
    10.0,  0.0, 0.0,  // 3
     9.0,  1.0, 0.0,  // 4
    10.0, 10.0, 0.0,  // 5
     9.0,  9.0, 0.0,  // 6
     0.0, 10.0, 0.0,  // 7
     1.0,  9.0, 0.0   // 8
};
const dTriIndex WhiteLine::Indices[IndexCount] = {
    1, 2, 3,
    2, 4, 3,
    3, 4, 5,
    4, 6, 5,
    6, 7, 5,
    6, 8, 7,
    7, 8, 1,
    8, 2, 1
};
////////////////////////////////////////////////////////////


// Փˌop֐
static void nearCallback( void *data, dGeomID o1, dGeomID o2 )
{
    static const int MAX_CONTACTS = 4; // ő̏Փˌo\

    dBodyID b1 = dGeomGetBody( o1 ); // 1
    dBodyID b2 = dGeomGetBody( o2 ); // 2

    if ( b1 && b2 && dAreConnectedExcluding( b1, b2, dJointTypeContact ) )
    return; // ՓˑΏۂłȂ̂̏Փ˂͂͂


    /////////////////////////////////////////////////////////////////////////
    // ZT[擾
    if (sensor_l) {
        if (o1 == sensor_l->geom || o2 == sensor_l->geom) {
            dContactGeom c;
            int numc = dCollide( o1, o2, 1, &c, sizeof(dContactGeom) );
            if (numc > 0) {
                if (o1 == w_line->geom || o2 == w_line->geom) {
                    hit_l = 1;
                    printf( "left sensor hit!! %f\n", c.depth );
                }
            }
            return;
        }
    }
    if (sensor_r) {
        if (o1 == sensor_r->geom || o2 == sensor_r->geom) {
            dContactGeom c;
            int numc = dCollide( o1, o2, 1, &c, sizeof(dContactGeom) );
            if (numc > 0) {
                if (o1 == w_line->geom || o2 == w_line->geom) {
                    hit_r = 1;
                    printf( "right sensor hit!! %f\n", c.depth );
                }
            }
            return;
        }
    }
    /////////////////////////////////////////////////////////////////////////

    dContact contact[MAX_CONTACTS];
    for ( int i=0; i<MAX_CONTACTS; i++ )
    {
        // ̓m̐ڐG̃p[^ݒ
        contact[i].surface.mode
            = dContactBounce
            | dContactSoftERP
            | dContactSoftCFM
            | dContactMu2
            | dContactSlip1
            | dContactSlip2;
        contact[i].surface.mu  = 300.0;        // CW
        contact[i].surface.mu2 = 300.0;        // CW2
        contact[i].surface.bounce = 0.5;       // W
        contact[i].surface.soft_erp = 0.3;     // ERPݒ
        contact[i].surface.soft_cfm = 0.0001;  // CFMݒ
        contact[i].surface.slip1 = 0.01;
        contact[i].surface.slip2 = 0.01;
    }

    // Փˌo
    int numc = dCollide( o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof( dContact ) ); 
    if ( numc > 0 )
    {
        for ( int i=0; i<numc; i++ )
        {
            dJointID c = dJointCreateContact( world, contactgroup, contact+i );
            dJointAttach( c, b1, b2 );
        }
    }
}

// simulation loop
static void simLoop( int pause )
{
    hit_r = 0;
    hit_l = 0;

    // Ctl+p ꂽifɓȂ
    if (!pause)
    {
        dSpaceCollide( space, 0, &nearCallback ); // Փˌo

        dWorldStep( world, 0.01 );
        
        dJointGroupEmpty( contactgroup );
        
        ///////////////////////////////////////////////////////////////////////////
        // [^
        if (hit_l == 0) {
            dJointSetHingeParam( dBodyGetJoint(car_wheel_l->body, 0), dParamVel, 10.0 );
        } else {
            dJointSetHingeParam( dBodyGetJoint(car_wheel_l->body, 0), dParamVel, -5.0 );
        }
        if (hit_r == 0) {
            dJointSetHingeParam( dBodyGetJoint(car_wheel_r->body, 0), dParamVel, 11.0 );
        } else {
            dJointSetHingeParam( dBodyGetJoint(car_wheel_r->body, 0), dParamVel, -5.0 );
        }
        ///////////////////////////////////////////////////////////////////////////
    }

 
    // draw object
    {
        std::vector<ST_Object*>::iterator it = drawObjList.begin();
        while (it != drawObjList.end())
        {
            ST_Object *obj = (*it);
            drawObj( obj );
            ++it;
        }
    }
    //// draw joint (debug)
    //{
    //    std::vector<ST_Object*>::iterator it = drawObjList.begin();
    //    while (it != drawObjList.end())
    //    {
    //        ST_Object *obj = (*it);
    //        drawJoint( obj );
    //        ++it;
    //    }
    //}
}
 
 
// start simulation - set viewpoint
static void start()
{
  static float xyz[3] = { -4.f, -4.f, 3.f };
  static float hpr[3] = { 45.f, -15.f, 0.f };
 
  dsSetViewpoint( xyz, hpr );
}


void createLineTracer()
{
    
    // car_body creating
    {
        car_body = new ST_Object;

        dReal pos[3] = { 0.0, 0.0, 1.0 };
        dReal size[3] = { 2.0, 1.0, 0.5 };

        // body setting
        car_body->body = dBodyCreate( world );
        dBodySetPosition( car_body->body, pos[0], pos[1], pos[2] );
        // geom setting
        car_body->geom = dCreateBox( space, size[0], size[1], size[2] );
        dGeomSetBody( car_body->geom, car_body->body );
        // mass setting
        dMass mass;
        dMassSetBoxTotal( &mass, 10.0, size[0], size[1], size[2] ); // 10kg
        dBodySetMass( car_body->body, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 0.0*M_PI/180 );
        dBodySetRotation( car_body->body, R );

        car_body->setColor( 0.0, 1.0, 1.0, 0.5 );


    }
    // car - front wheel creating
    {
        car_wheel_f = new ST_Object;

        dReal pos[3] = { 1.0, 0.0, 0.2 };
        dReal radius = 0.2;

        // body setting
        car_wheel_f->body = dBodyCreate( world );
        dBodySetPosition( car_wheel_f->body, pos[0], pos[1], pos[2] );
        // geom setting
        car_wheel_f->geom = dCreateSphere( space, radius );
        dGeomSetBody( car_wheel_f->geom, car_wheel_f->body );
        // mass setting
        dMass mass;
        dMassSetSphereTotal( &mass, 0.5, radius );
        dBodySetMass( car_wheel_f->body, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 0.0*M_PI/180 );
        dBodySetRotation( car_wheel_f->body, R );

        car_wheel_f->setColor( 0.3, 0.3, 0.3 );


        // joint creating
        dJointID w_axis = dJointCreateBall( world, 0 );
        dJointAttach( w_axis, car_body->body, car_wheel_f->body );
        dJointSetBallAnchor( w_axis, pos[0], pos[1], pos[2] );
    }
    // car - left wheel creating
    {
        car_wheel_l = new ST_Object;

        dReal pos[3] = { -1.0, 1.0, 0.2 };
        dReal radius = 0.2;
        dReal length = 0.3;

        // body setting
        car_wheel_l->body = dBodyCreate( world );
        dBodySetPosition( car_wheel_l->body, pos[0], pos[1], pos[2] );
        // geom setting
        car_wheel_l->geom = dCreateCylinder( space, radius, length );
        dGeomSetBody( car_wheel_l->geom, car_wheel_l->body );
        // mass setting
        dMass mass;
        dMassSetCylinderTotal( &mass, 0.5, 1, radius, length );
        dBodySetMass( car_wheel_l->body, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 90.0*M_PI/180 );
        dBodySetRotation( car_wheel_l->body, R );

        car_wheel_l->setColor( 0.3, 0.3, 0.3 );

        // joint creating
        dJointID w_axis = dJointCreateHinge( world, 0 );
        dJointAttach( w_axis, car_body->body, car_wheel_l->body );
        dJointSetHingeAnchor( w_axis, pos[0], pos[1], pos[2] );
        dJointSetHingeAxis( w_axis, 0.0, -1.0, 0.0 );
        
        dJointSetHingeParam( w_axis, dParamFMax, 10.0 );
    }
    // car - right wheel creating
    {
        car_wheel_r = new ST_Object;

        dReal pos[3] = { -1.0, -1.0, 0.2 };
        dReal radius = 0.2;
        dReal length = 0.3;

        // body setting
        car_wheel_r->body = dBodyCreate( world );
        dBodySetPosition( car_wheel_r->body, pos[0], pos[1], pos[2] );
        // geom setting
        car_wheel_r->geom = dCreateCylinder( space, radius, length );
        dGeomSetBody( car_wheel_r->geom, car_wheel_r->body );
        // mass setting
        dMass mass;
        dMassSetCylinderTotal( &mass, 0.5, 1, radius, length );
        dBodySetMass( car_wheel_r->body, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 90.0*M_PI/180 );
        dBodySetRotation( car_wheel_r->body, R );

        car_wheel_r->setColor( 0.3, 0.3, 0.3 );


        // joint creating
        dJointID w_axis = dJointCreateHinge( world, 0 );
        dJointAttach( w_axis, car_body->body, car_wheel_r->body );
        dJointSetHingeAnchor( w_axis, pos[0], pos[1], pos[2] );
        dJointSetHingeAxis( w_axis, 0.0, -1.0, 0.0 );

        dJointSetHingeParam( w_axis, dParamFMax, 10.0 );
    }

    // car sensor body creating
    {
        car_sensor = new ST_Object;

        dReal pos[3] = { 0.0, 0.0, 0.2 };
        dReal size[3] = { 0.5, 2.5, 0.1 };

        // body setting
        car_sensor->body = dBodyCreate( world );
        dBodySetPosition( car_sensor->body, pos[0], pos[1], pos[2] );
        // geom setting
        car_sensor->geom = dCreateBox( space, size[0], size[1], size[2] );
        dGeomSetBody( car_sensor->geom, car_sensor->body );
        // mass setting
        dMass mass;
        dMassSetBoxTotal( &mass, 10.0, size[0], size[1], size[2] ); // 10kg
        dBodySetMass( car_sensor->body, &mass );
        // rotation
        dMatrix3 R;
        dRFromAxisAndAngle( R, 1.0, 0.0, 0.0, 0.0*M_PI/180 );
        dBodySetRotation( car_sensor->body, R );

        car_sensor->setColor( 1.0, 1.0, 0.0, 0.5 );


        // joint creating
        dJointID fixed = dJointCreateFixed( world, 0 );
        dJointAttach( fixed, car_body->body, car_sensor->body );
        dJointSetFixed( fixed );
    }

    // left sensor setting 
    {
        sensor_l = new ST_Object;

        dReal pos[3] = { 0.0, 1.0, 0.0 };
        dReal length = 0.5;

        sensor_l->geom = dCreateRay( space, length );
        dGeomSetBody( sensor_l->geom, car_sensor->body );

        // position and rotation setting
        dGeomSetOffsetPosition( sensor_l->geom, pos[0], pos[1], pos[2] );
        dMatrix3 R;
        dRFromAxisAndAngle( R, 0.0, 1.0, 0.0, 180.0*M_PI/180 );
        dGeomSetOffsetRotation( sensor_l->geom, R );
        //dGeomRaySet( sensor_l->geom, pos[0], pos[1], pos[2], dir[0], dir[1], dir[2] );

        sensor_l->setColor( 1.0, 0.0, 0.0 );
    }
    // right sensor setting 
    {
        sensor_r = new ST_Object;

        dReal pos[3] = { 0.0, -1.0, 0.0 };
        dReal length = 0.5;

        sensor_r->geom = dCreateRay( space, length );
        dGeomSetBody( sensor_r->geom, car_sensor->body );

        // position and rotation setting
        dGeomSetOffsetPosition( sensor_r->geom, pos[0], pos[1], pos[2] );
        dMatrix3 R;
        dRFromAxisAndAngle( R, 0.0, 1.0, 0.0, 180.0*M_PI/180 );
        dGeomSetOffsetRotation( sensor_r->geom, R );
        //dGeomRaySet( sensor_r->geom, pos[0], pos[1], pos[2], dir[0], dir[1], dir[2] );

        sensor_r->setColor( 0.0, 1.0, 0.0 );
    }

    drawObjList.push_back( car_wheel_f );
    drawObjList.push_back( car_wheel_l );
    drawObjList.push_back( car_wheel_r ); 
    drawObjList.push_back( sensor_l );
    drawObjList.push_back( sensor_r ); 
    drawObjList.push_back( car_sensor );
    drawObjList.push_back( car_body );
}

void createWhiteLine()
{
    // 
    {
        w_line = new ST_Object;

        dTriMeshDataID data;
        data = dGeomTriMeshDataCreate();
        dGeomTriMeshDataBuildSingle(
            data, WhiteLine::Vertices, 3*sizeof(float), WhiteLine::VertexCount,
            WhiteLine::Indices, WhiteLine::IndexCount, 3*sizeof(dTriIndex));
     
        w_line->geom = dCreateTriMesh( space, data, 0, 0, 0 );
        dGeomSetPosition( w_line->geom, -1.5, -0.5, 0.01 );
        dGeomSetData( w_line->geom, data );

        w_line->setColor( 1.0, 1.0, 1.0 );
    }

    drawObjList.push_back( w_line );
}


 
int main( int argc, char* argv[] )
{
    dInitODE();
 
    // setup pointers to drawstuff callback functions
    dsFunctions fn;
    fn.version = DS_VERSION;
    fn.start   = &start;
    fn.step    = &simLoop;
    fn.command = 0;
    fn.stop    = 0;
    fn.path_to_textures = "../drawstuff/textures";
 
    world = dWorldCreate();
    dWorldSetGravity( world, 0.0, 0.0, -9.8 );

    space = dHashSpaceCreate( 0 ); 
    contactgroup = dJointGroupCreate( 0 );

    dCreatePlane( space, 0, 0, 1, 0 );

    createLineTracer();  // ԑ̂쐬
    createWhiteLine();   // 

    dsSimulationLoop( argc, argv, 320, 240, &fn );

    dWorldDestroy( world );
    dCloseODE();
    return 0;
}



